/* -*- MaTX -*-
 *
 * halist.mm: an `h' associated list package
 *
 *  Originally written by KIYOTA Hiromitsu <1999/04/15>
 *  Copyright (c) 1999 Hiromitsu Kiyota, kiyota@ds.mei.titech.ac.jp
 *
 * Usage:
 *  List halist(id, pairs) List id, pairs;
 *    id: ex. {"alist1"}, {1}, etc., but
 *        Lists of only null-lists are reserved for the internals.
 *   halist({id, method}) String method;
 *     method: "exists", "length", "clear", "all", "keys" or "values" for id
 *   halist({id, "access"}, {args}) == halist(id, {args})
 *     args: {key1, val1}, {key2, val2}, ... to add/change those pairs
 *       String key1;  (Any type) val2
 *     args: key1, key2, ... to return a list of the corresponding values
 *   halist({id, "delete"}, {args})
 *     args: key1, key2, ... to delete the corresponding pairs
 *
 * RCS: $Id: halist.mm,v 1.4 1999/05/05 14:38:37 kiyota Exp $
 */

#ifndef _halist_mm
#define _halist_mm

//#define _halist_listeq_func
#define _halist_listeq_undef_hack
//#define _halist_listeq_debug
//#define _halist_list_workaround

#ifdef _halist_listeq_func
Func Integer
_halist_listeq(a, b)
     List a, b;
{
  Integer s, t, i;

#ifdef _halist_listeq_undef_hack  
  static Integer undef_t;

  if (undef_t == 0)  // when first called
    {
      undef_t = typeof(makelist(1), 1);  // makes UNDEF
    }
#endif /* _halist_listeq_undef_hack */

#ifdef _halist_listeq_debug
  print a,b;
#endif /* _halist_listeq_debug */

  s = length(a);
#ifdef _halist_listeq_debug
  print s;
#endif /* _halist_listeq_debug */
  if (s != length(b))
    {
      return 0;
    }
  else
    {
      for (i = 1; i <= s; ++i)
	{
	  t = typeof(a, i);
#ifdef _halist_listeq_debug
	  print t;
#endif /* _halist_listeq_debug */
	  if (t != typeof(b, i))
	    {
	      return 0;
	    }
	  if (t == List && ! _halist_listeq(a(i, List), b(i, List))
	      || t == Integer && a(i, Integer) != b(i, Integer)
	      || t == Real && a(i, Real) != b(i, Real)
	      || t == Complex && a(i, Complex) != b(i, Complex)
	      || t == String && a(i, String) != b(i, String)
	      || t == Polynomial && a(i, Polynomial) != b(i, Polynomial)
	      || t == Rational && a(i, Rational) != b(i, Rational)
	      || t == Matrix && a(i, Matrix) != b(i, Matrix)
	      || t == Array && Matrix(a(i, Array)) != Matrix(b(i, Array))
//	      || t == Array && a(i, Array) != b(i, Array)  // it errors!
	      || t == Index && a(i, Index) != b(i, Index)
#ifdef _halist_listeq_undef_hack
	      || t == undef_t
#else /* ! _halist_listeq_undef_hack */
	      || t == Undef  // consistent, or not??
#endif /* _halist_listeq_undef_hack */
	    )  // Oops!
	    {
	      return 0;
	    }
	}
    }
  return 1;
}
#endif /* _halist_listeq_func */


// To convert List-boolian {}/{{}} to Integer-boolian 0/1
Func Integer
_halist_listbool(l, n, ...)
     List l;
     Integer n;
{
  switch (nargs)
    {
    case 1:  // the object to convert is the list itself
      return length(l);
      break;
    case 2:  // the object(s) to convert is a list element of the list l
      if (n <= 0 || n > length(l))
	{
	  error("_halist_listzerop(): Invalid argument. (n=%d)\n", n);
	}
      else if (typeof(l, n) != List)
	{
	  error("_halist_listzerop(): Invalid element of the list. (n=%d)\n", n);
	}
      return length(l(n, List));
      break;
    default:
      error("_halist_listzerop(): Incorrect number of arguments. (nargs=%d)\n", nargs);
      break;
    }
}	  


Func List
halist(id, args, ...)
     List id;  // user-defined id must not be list-defined natural numbers.
     List args;
{
  static List n;  // defaults to {} (== 0)
  String cmd;
  static List l;  // {id1, {pairs1}, id2, {pairs2}, ...}
  Integer i, j, na;
  List s;
#ifndef _halist_list_workaround
  Index tm;  // matching index of keys in the alist and the given one
  Array td;  // for 0-1 array in delete
#else /* _halist_list_workaround */
  Integer k;
  List r, t, u;
#endif /* ! _halist_list_workaround */

  switch (nargs)  // terribly long switch statement 8-X
    {
    case 0:  // return a safe alist identifier
      do
	{
	  n = n + {{}};  // n >= 1
	}
      while (_halist_listbool(halist({n, "exists"})));  // terrible, isn't it?
      return n;
      break;

    case 1:  // provides method access for the alist, and
      // each returns a possibly non-ordered set of something
      if (length(id) == 2 && typeof(id, 1) == List && typeof(id, 2) == String)
	{
#ifndef _halist_list_workaround	  
	  {id, cmd} = id;  // depends on the length of id == 2
#else /* _halist_list_workaround */
	  cmd = id(2, String);
	  id = id(1, List);
#endif /* ! _halist_list_workaround */
	}
      else
	{
	  cmd = "all";
	}

      // select the alist of id
#ifndef _halist_list_workaround
      if (i = length(l))
	{
	  tm = find(l(Index([1:length(l)/2]*2) .- 1)
		    .== {id}*(length(l)/2))*2 .- 1;
	  if (i = length(tm))
	    {
	      i = tm(1);
	    }
	}
      // when first called or new id given, you come here with i == 0 ;)
#else /* _halist_list_workaround */
      for (i = length(l) - 1;
	   i >= 1
#ifndef _halist_listeq_func
	   && l(i, List) != id;
#else /* _halist_listeq_func */
	   && ! _halist_listeq(l(i, List), id);
#endif /* ! _halist_listeq_func */
	   i = i - 2) {}
#endif /* ! _halist_list_workaround */
      if (cmd == "exists")  // this is a CLASS method, maybe ;)
	{
	  if (i < 1)
	    {
	      return {};  /* 0 */
	    }
	  else
	    {
	      return {{}};  /* 1 */
	    }
	}
      if (i < 1)
	{
	  l = l + {id, {}};  // new alist
	  i = length(l) - 1;
	}
#ifdef _halist_list_workaround
      r = l(i + 1, List);  // the alist of id
#endif /* _halist_list_workaround */

      if (cmd == "length")
	{
#ifndef _halist_list_workaround
	  return {length(l(i + 1, List))};
#else /* _halist_list_workaround */
	  return {length(r)};
#endif /* ! _halist_list_workaround */
	}
      else if (cmd == "clear" || cmd == "new")  // different things? B)
	{
#ifndef _halist_list_workaround
	  return l(i + 1) = {};
#else /* _halist_list_workaround */
	  l(i + 1) = {};  // what is the value of this???
	  return {};  // of course, meaningful `0'
#endif /* ! _halist_list_workaround */
	}
      else if (cmd == "all")
	{
#ifndef _halist_list_workaround
	  return l(i + 1, List);
#else /* _halist_list_workaround */
	  return r;
#endif /* ! _halist_list_workaround */
	}
      else if (cmd == "keys")
	{
#ifndef _halist_list_workaround
	  return l(i + 1, Index([1:length(l(i + 1, List))]), 1);
#else /* _halist_list_workaround */
	  s = {};
	  for (j = length(r); j > 0; --j)
	    {
	      t = r(j, List);
	      s = s + t(1:1);  // {"key"} for now, but maybe any in future
	    }
	  return s;
#endif /* ! _halist_list_workaround */
	}
      else if (cmd == "values")
	{
#ifndef _halist_list_workaround
	  return l(i + 1, Index([1:length(l(i + 1, List))]), 2);
#else /* _halist_list_workaround */
	  s = {};
	  for (j = length(r); j > 0; --j)
	    {
	      t = r(j, List);
	      s = s + t(2:2);
	    }
	  return s;
#endif /* ! _halist_list_workaround */
	}
      else if (cmd == "print")
	{
#ifndef _halist_list_workaround
	  print l(i + 1, List);
#else /* _halist_list_workaround */
	  print r;
#endif /* ! _halist_list_workaround */
	  return {{}};  // possible, meaningless `1'
	}
      else if (cmd == "alprint")
	{
	  print l;
	  return {{}};  // possible, meaningless `1'
	}
      else
	{
	  error("halist(): Invalid method. (\"%s\")\n", cmd);
	}
      break;

    case 2:  // provides method access for pairs of the alist, and
      // each returns an ordered set of something
      if (length(id) == 2 && typeof(id, 1) == List && typeof(id, 2) == String)
	{
#ifndef _halist_list_workaround
	  {id, cmd} = id;  // depends on the length of id == 2
#else /* _halist_list_workaround */
	  cmd = id(2, String);
	  id = id(1, List);
#endif /* ! _halist_list_workaround */
	}
      else
	{
	  cmd = "";  // to what method defaults?
	}

      // select the alist of id
#ifndef _halist_list_workaround
      if (i = length(l))
	{
	  tm = find(l(Index([1:length(l)/2]*2) .- 1)
		    .== {id}*(length(l)/2))*2 .- 1;
	  if (i = length(tm))
	    {
	      i = tm(1);
	    }
	}
      // when first called or new id given, you come here with i == 0 ;)
#else /* _halist_list_workaround */
      for (i = length(l) - 1;
	   i >= 1
#ifndef _halist_listeq_func
	   && l(i, List) != id;
#else /* _halist_listeq_func */
	   && ! _halist_listeq(l(i, List), id);
#endif /* ! _halist_listeq_func */
	   i = i - 2) {}
#endif /* ! _halist_list_workaround */
      if (i < 1)
	{
	  print id;
	  error("halist(): Incorrect identifier.\n");
	}
#ifdef _halist_list_workaround
      r = l(i + 1, List);  // the alist of id
#endif /* _halist_list_workaround */
      s = {};  // return list to be

      if (cmd == "" || cmd == "access" /* etc. */)
	{
	  // when args == {}, do nothing
	  for ({na = length(args), j = 1}; j <= na; ++j)
	    {
	      switch (typeof(args, j))
		{
		case List:  // mathod: "", "access", ("set",) etc.
#ifndef _halist_list_workaround
		  if (length(args(j, List)) != 2
		      || typeof(args(j, 1), 1) != String)
#else /* _halist_list_workaround */
		  t = args(j, List);  // a pair of key and value of args
		  if (length(t) != 2 || typeof(t, 1) != String)
#endif /* ! _halist_list_workaround */
		    {
#ifndef _halist_list_workaround
		      print args(j, List);
#else /* _halist_list_workaround */
		      print t;
#endif /* ! _halist_list_workaround */
		      error("halist(): Invalid pair of key and value.\n");
		    }
#ifndef _halist_list_workaround
		  if (length(l(i + 1, List))
		      && length(tm = find(l(i + 1,
					    Index([1:length(l(i + 1, List))]),
					    1)
					  .== args(j, 1, String))) > 0)
		    {
		      l(i + 1, tm(1)) = args(j, List);
		    }
		  else
		    {
		      l(i + 1) = l(i + 1, List) + {args(j, List)};  // new pair
		    }
		  s = s + args(j, 2);
#else /* _halist_list_workaround */
		  for (k = length(r); k > 0; --k)
		    {
		      u = r(k, List);  // a pair of key and value of the alist
		      if (u(1, String) == t(1, String))
			{
			  r(k) = t;
			  break;
			}
		    }
		  if (k == 0)
		    {
		      r = r + {t};  // new pair
		    }
		  s = s + t(2:2);
#endif /* ! _halist_list_workaround */
		  break;
		case String:  // method: "", "access", ("get",) etc.
#ifndef _halist_list_workaround
		  if (length(l(i + 1, List))
		      && length(tm = find(l(i + 1,
					    Index([1:length(l(i + 1, List))]),
					    1)
					  .== args(j, String))) > 0)
		    {
		      s = s + l(i + 1, tm(1), 2);
		    }
		  else
#else /* _halist_list_workaround */
		  for (k = length(r); k > 0; --k)
		    {
		      u = r(k, List);  // a pair of key and value of the alist
		      if (u(1, String) == args(j, String))
			{
			  s = s + u(2:2);
			  break;
			}
		    }
		  if (k == 0)
#endif /* ! _halist_list_workaround */
		    {
		      error("halist(): Invalid key. (\"%s\")\n", args(j, String));
		    }
		  break;
		default:
		  print args;
		  error("halist(): Invalid type of %d-th element of args.\n", length(args) - j + 1);
		  break;
		}
	    }
	}
      else if (cmd == "delete")
	{
#ifndef _halist_list_workaround
	  td = ONE(1, length(l(i + 1, List)));
#endif /* ! _halist_list_workaround */
	  // when args == {}, do nothing
	  for ({na = length(args), j = 1}; j <= na; ++j)
	    {
	      if (typeof(args, j) != String)
		{
		  print args;
		  error("halist(): Invalid type of %d-th element of args.\n", j);
		}
#ifndef _halist_list_workaround
	      if (length(l(i + 1, List))
		  && length(tm = find(l(i + 1,
					Index([1:length(l(i + 1, List))]),
					1)
				      .== args(j, String))) > 0)
		{
		  s = s + l(i + 1, tm(1), 2);
		  td(tm(1)) = 0;
		}
	      else
#else /* _halist_list_workaround */
	      for (k = length(r); k > 0; --k)
		{
		  u = r(k, List);  // a pair of key and value of the alist
		  if (length(u)  // skip already deleted pairs
		      && u(1, String) == args(j, String))
		    {
		      s = s + u(2:2);
		      r(k) = {};
		      break;
		    }
		}
	      if (k == 0)
#endif /* ! _halist_list_workaround */
		{
		  error("halist(): Invalid key. (\"%s\")\n", args(j, String));
		}
	    }
	  // normalization of the alist
#ifndef _halist_list_workaround
	  l(i + 1) = l(i + 1, td);
#else /* _halist_list_workaround */
	  t = {};
	  for (k = length(r); k > 0; --k)
	    {
	      if (length(r(k, List)))  // excludes {}
		{
		  t = t + r(k:k);
		}
	    }
	  r = t;
#endif /* ! _halist_list_workaround */
	}
      else if (cmd == "exists")
	{
	  // when args == {}, this equals to halist({id, "exists"})
	  na = length(args);
	  if (na == 0)
	    {
	      return halist({id, "exists"});
	    }
	  for (j = 1; j <= na; ++j)
	    {
	      if (typeof(args, j) != String)
		{
		  print args;
		  error("halist(): Invalid type of %d-th element of args.\n", j);
		}
#ifndef _halist_list_workaround
	      if (length(l(i + 1, List))
		  && any(l(i + 1, Index([1:length(l(i + 1, List))]), 1)
			 .== args(j, String)))  // exists
#else /* _halist_list_workaround */
	      for (k = length(r); k > 0; --k)
		{
		  u = r(k, List);  // a pair of key and value of the alist
		  if (u(1, String) == args(j, String))
		    {
		      break;
		    }
		}
	      if (k != 0)  // exists
#endif /* ! _halist_list_workaround */
		{
		  s = s + {{{}}};  /* 1 for _halist_listbool() */
		}
	      else
		{
		  s = s + {{}};  /* 0 for _halist_listbool() */
		}
	    }
	}
      else
	{
	  error("halist(): Invalid method. (\"%s\")\n", cmd);
	}

#ifdef _halist_list_workaround
      l(i + 1) = r;
#endif /* _halist_list_workaround */
      return s;
      break;

    default:
      error("halist(): Incorrect number of arguments. (nargs=%d)\n", nargs);
      break;
    }
}

#endif /* ! _halist_mm */
